@@ -1,4 +1,5 @@
 /*
+ *  Copyright (c) 2014, Peter Haag
  *  Copyright (c) 2009, Peter Haag
  *  Copyright (c) 2004-2008, SWITCH - Teleinformatikdienste fuer Lehre und Forschung
  *  All rights reserved.
@@ -11,7 +12,7 @@
  *   * Redistributions in binary form must reproduce the above copyright notice, 
  *     this list of conditions and the following disclaimer in the documentation 
  *     and/or other materials provided with the distribution.
- *   * Neither the name of SWITCH nor the names of its contributors may be 
+ *   * Neither the name of the author nor the names of its contributors may be 
  *     used to endorse or promote products derived from this software without 
  *     specific prior written permission.
  *  
@@ -44,6 +45,7 @@
 #include <sys/stat.h>
 #include <sys/uio.h>
 #include <unistd.h>
+#include <signal.h>
 #include <fcntl.h>
 #include <errno.h>
 #include <ctype.h>
@@ -69,6 +71,8 @@
 #include "nfstatfile.h"
 #include "expire.h"
 
+static uint32_t timeout = 0;
+
 /* 
  * expire.c is needed for daemon code as well as normal stdio code 
  * therefore a generic LogError is defined, which maps to the 
@@ -89,6 +93,41 @@ static int unlink_debug (const char *path) {
 } // End of unlink_debug
 #endif
 
+static void IntHandler(int signal) {
+
+	switch (signal) {
+		case SIGALRM:
+			timeout = 1;
+			break;
+		case SIGHUP:
+		case SIGINT:
+		case SIGTERM:
+			timeout = 1;
+			break;
+			break;
+		case SIGCHLD:
+		default:
+			// ignore everything we don't know
+			break;
+	}
+
+} /* End of IntHandler */
+
+static void SetupSignalHandler(void) {
+struct sigaction act;
+
+	memset((void *)&act,0,sizeof(struct sigaction));
+	act.sa_handler = IntHandler;
+	sigemptyset(&act.sa_mask);
+	act.sa_flags = 0;
+	sigaction(SIGTERM, &act, NULL);
+	sigaction(SIGINT, &act, NULL);
+	sigaction(SIGHUP, &act, NULL);
+	sigaction(SIGALRM, &act, NULL);
+	sigaction(SIGCHLD, &act, NULL);
+
+} // End of SetupSignalHandler
+
 uint64_t ParseSizeDef(char *s, uint64_t *value) {
 char *p;
 uint64_t fac;
@@ -318,7 +357,7 @@ char		first_timestring[16], last_timestring[16];
 
 } // End of RescanDir
 
-void ExpireDir(char *dir, dirstat_t *dirstat, uint64_t maxsize, uint64_t maxlife ) {
+void ExpireDir(char *dir, dirstat_t *dirstat, uint64_t maxsize, uint64_t maxlife, uint32_t runtime ) {
 FTS 		*fts;
 FTSENT 		*ftsent;
 uint64_t	sizelimit, num_expired;
@@ -331,6 +370,11 @@ time_t 		now = time(NULL);
 	if ( dirstat->low_water == 0 )
 		dirstat->low_water = 95;
 
+	if ( runtime ) {
+		SetupSignalHandler();
+		alarm(runtime);
+	}
+
 	if ( maxlife ) {
 		// build an appropriate string for comparing
 		time_t t_expire = now - maxlife;
@@ -432,7 +476,7 @@ time_t 		now = time(NULL);
 						lifetime_done = 1;
 					}
 				}
-				done = size_done && lifetime_done;
+				done = (size_done && lifetime_done) || timeout;
 
 			}
 		} else {
@@ -467,6 +511,11 @@ time_t 		now = time(NULL);
 		// the time limit is shorter than the latest file
 		dirstat->first = dirstat->last;
 	}
+	if ( runtime )
+		alarm(0);
+	if ( timeout ) {
+		LogError( "Maximum execution time reached! Interrupt expire.\n");
+	}
 	if ( num_expired > dirstat->numfiles ) {
 		LogError( "Error updating stat record: Number of files inconsistent!\n");
 		LogError( "Will automatically rebuild this directory next time\n");
@@ -552,7 +601,7 @@ channel_t *current_channel = channel;
 
 } // End of PrepareDirLists
 
-void ExpireProfile(channel_t *channel, dirstat_t *current_stat, uint64_t maxsize, uint64_t maxlife ) {
+void ExpireProfile(channel_t *channel, dirstat_t *current_stat, uint64_t maxsize, uint64_t maxlife, uint32_t runtime ) {
 int  		size_done, lifetime_done, done;
 char 		*expire_timelimit = "";
 time_t		now = time(NULL);
@@ -562,6 +611,7 @@ uint64_t	sizelimit, num_expired;
 		return;
 
 	done = 0;
+	SetupSignalHandler();
 
 	if ( maxlife ) {
 //		time_t t_expire = now - maxlife;
@@ -582,6 +632,8 @@ uint64_t	sizelimit, num_expired;
 	num_expired = 0;
 
 	PrepareDirLists(channel);
+	if ( runtime )
+		alarm(runtime);
 	while ( !done ) {
 		char *p;
 		int file_removed;
@@ -678,6 +730,8 @@ uint64_t	sizelimit, num_expired;
 		} else 
 			// all done
 			done = 1;
+		if ( timeout ) 
+			done = 1;
 
 		// advance fts entry in expire channel to next file, if file was removed
 		if ( file_removed ) {
@@ -737,6 +791,12 @@ uint64_t	sizelimit, num_expired;
 		}
 	} // while ( !done )
 
+	if ( runtime )
+		alarm(0);
+	if ( timeout ) {
+		LogError( "Maximum execution time reached! Interrupt expire.\n");
+	}
+
 } // End of ExpireProfile
 
 void UpdateBookStat(dirstat_t *dirstat, bookkeeper_t *books) {
