@@ -1,4 +1,5 @@
 /*
+ *  Copyright (c) 2014, Peter Haag
  *  Copyright (c) 2009, Peter Haag
  *  Copyright (c) 2008, SWITCH - Teleinformatikdienste fuer Lehre und Forschung
  *  All rights reserved.
@@ -11,7 +12,7 @@
  *   * Redistributions in binary form must reproduce the above copyright notice, 
  *     this list of conditions and the following disclaimer in the documentation 
  *     and/or other materials provided with the distribution.
- *   * Neither the name of SWITCH nor the names of its contributors may be 
+ *   * Neither the name of the author nor the names of its contributors may be 
  *     used to endorse or promote products derived from this software without 
  *     specific prior written permission.
  *  
@@ -85,6 +86,10 @@
  *   +-----------+-------------+-------------+-------------+-----+-------------+
  */
 
+#define NOT_COMPRESSED 0
+#define LZO_COMPRESSED 1
+#define BZ2_COMPRESSED 2
+
 typedef struct file_header_s {
 	uint16_t	magic;				// magic to recognize nfdump file type and endian type
 #define MAGIC 0xA50C
@@ -93,14 +98,13 @@ typedef struct file_header_s {
 #define LAYOUT_VERSION_1	1
 
 	uint32_t	flags;				
-#define NUM_FLAGS		3
-#define FLAG_COMPRESSED 	0x1		// flow records are compressed
+#define NUM_FLAGS		4
+#define FLAG_NOT_COMPRESSED	0x0		// records are not compressed
+#define FLAG_LZO_COMPRESSED	0x1		// records are LZO compressed
 #define FLAG_ANONYMIZED 	0x2		// flow data are anonimized 
 #define FLAG_CATALOG		0x4		// has a file catalog record after stat record
-
-									/*
-										0x1 File is compressed with LZO1X-1 compression
-									 */
+#define FLAG_BZ2_COMPRESSED 0x8		// records are BZ2 compressed
+							
 	uint32_t	NumBlocks;			// number of data blocks in file
 	char		ident[IDENTLEN];	// string identifier for this file
 } file_header_t;
@@ -163,7 +167,7 @@ typedef struct data_block_header_s {
 	uint32_t	NumRecords;		// number of data records in data block
 	uint32_t	size;			// size of this block in bytes without this header
 	uint16_t	id;				// Block ID == DATA_BLOCK_TYPE_2
-	uint16_t	flags;			// 0 - kompatibility
+	uint16_t	flags;			// 0 - compatibility
 								// 1 - block uncompressed
 								// 2 - block compressed
 } data_block_header_t;
@@ -255,18 +259,24 @@ typedef struct nffile_s {
  * Type 4: xstat - bpp histogram record
  */
 
-#define CommonRecordType	1
+#define CommonRecordV0Type	1
 #define ExtensionMapType	2
 #define PortHistogramType	3
 #define BppHistogramType	4
-// TC code
+
+// TC code - phased out
 #define ExporterRecordType	5
 #define SamplerRecordype	6
 
+// replaces TC Types
 #define ExporterInfoRecordType	7
 #define ExporterStatRecordType	8
 #define SamplerInfoRecordype	9
 
+// new extended Common Record as intermediate solution to overcome 255 exporters
+// requires moderate changes till 1.7
+#define CommonRecordType	10
+
  /* 
  * All records are 32bit aligned and layouted in a 64bit array. The numbers placed in () refer to the netflow v9 type id.
  *
@@ -283,10 +293,9 @@ typedef struct nffile_s {
  * bit  2:	0: 32bit dOctets	 1: 64bit dOctets 
  * bit  3:  0: IPv4 next hop     1: IPv6 next hop
  * bit  4:  0: IPv4 BGP next hop 1: BGP IPv6 next hop
- * bit  5:  0:                   1:
- * bit  6:  0:                   1:
- * bit  7:  0:                   1:
- * bit  8:  0: unsampled         1: sampled flow - sampling applied
+ * bit  5:  0: IPv4 exporter IP  1: IPv6 exporter IP
+ * bit  6:  0: flow              1: event
+ * bit  7:  0: unsampled         1: sampled flow - sampling applied
  * 
  * Required extensions: 1,2,3
  * ------------------------------
@@ -297,7 +306,7 @@ typedef struct nffile_s {
  * Extension 3: 32 or 64 bit byte counter           Flags bit 2: 0: 32bit, 1: 64bit
  * 
  * Commmon record - extension 0
- * *+---+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
  * |  - |       0      |      1       |      2       |      3       |      4       |      5       |      6       |      7       |
  * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
  * |  0 |         record type == 1    |             size            |    flags     |    tag       |           ext. map          |
@@ -308,6 +317,20 @@ typedef struct nffile_s {
  * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
  * |  3 |           srcport (7)       |   dstport(11)/ICMP (32)     |
  * +----+--------------+--------------+--------------+--------------+
+ *
+ * Commmon record - extension 0 - Type 10
+ * required for larger exporter ID reference 
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  - |       0      |      1       |      2       |      3       |      4       |      5       |      6       |      7       |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  0 |         record type == 10   |             size            |            flags            |           ext. map          |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  1 |          msec_first         |           msec_last         |                          first (22)                       |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  2 |                          last (21)                        |fwd_status(89)| tcpflags (6) |  proto (4)   |  src tos (5) |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  3 |           srcport (7)       |   dstport(11)/ICMP (32)     |          exporter ID        |           <free>            |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
 
  * 
  */
@@ -326,19 +349,49 @@ typedef struct common_record_s {
  	uint16_t	size;
 
 	// record meta data
-	uint8_t		flags;
+	uint16_t	flags;
 #define FLAG_IPV6_ADDR	1
 #define FLAG_PKG_64		2
 #define FLAG_BYTES_64	4
 #define FLAG_IPV6_NH	8
 #define FLAG_IPV6_NHB	16
 #define FLAG_IPV6_EXP	32
+#define FLAG_EVENT		64
 #define FLAG_SAMPLED	128
 
 #define SetFlag(var, flag) 		(var |= flag)
 #define ClearFlag(var, flag) 	(var &= ~flag)
 #define TestFlag(var, flag)		(var & flag)
 
+ 	uint16_t	ext_map;
+
+	// netflow common record
+ 	uint16_t	msec_first;
+ 	uint16_t	msec_last;
+ 	uint32_t	first;
+ 	uint32_t	last;
+ 
+ 	uint8_t		fwd_status;
+ 	uint8_t		tcp_flags;
+ 	uint8_t		prot;
+ 	uint8_t		tos;
+ 	uint16_t	srcport;
+ 	uint16_t	dstport;
+
+	uint16_t	exporter_sysid;
+	uint16_t	reserved;
+
+	// link to extensions
+ 	uint32_t	data[1];
+} common_record_t;
+
+typedef struct common_record_v0_s {
+ 	// record head
+ 	uint16_t	type;
+ 	uint16_t	size;
+
+	// record meta data
+	uint8_t		flags;
 	uint8_t		exporter_sysid;
  	uint16_t	ext_map;
 
@@ -357,8 +410,10 @@ typedef struct common_record_s {
 
 	// link to extensions
  	uint32_t	data[1];
-} common_record_t;
-#define COMMON_RECORD_DATA_SIZE (sizeof(common_record_t) - sizeof(uint32_t) )
+} common_record_v0_t;
+
+#define COMMON_RECORD_DATA_SIZE   (sizeof(common_record_t) - sizeof(uint32_t) )
+#define COMMON_RECORDV0_DATA_SIZE (sizeof(common_record_v0_t) - sizeof(uint32_t) )
 
 #define COMMON_BLOCK	0
 
@@ -420,6 +475,7 @@ typedef struct ip_addr_s {
 		};
 		uint64_t		_v6[2];
 	} ip_union;
+#define IP_ADDR_T
 } ip_addr_t;
 
 #define v4 ip_union._v4
@@ -915,15 +971,145 @@ typedef struct tpl_ext_27_s {
 #define EX_RESERVED_8	35
 #define EX_RESERVED_9	36
 
-#define EX_NSEL_0		37
-#define EX_NSEL_1		38
-#define EX_NSEL_2		39
-#define EX_NSEL_3		40
-#define EX_NSEL_4		41
+/*
+ * NSEL Common block
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  0 |                                           NF_F_FLOW_CREATE_TIME_MSEC(152)                                             |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  1 |                      NF_F_CONN_ID(148)                    |i type(176/8) |i code(177/9) |EVT(40005/233)|    fill      |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  2 |   NF_F_FW_EXT_EVENT(33002)  |            fill2            |
+ * +----+--------------+--------------+--------------+--------------+
+ * * EVT: NF_F_FW_EVENT
+ * * XEVT: NF_F_FW_EXT_EVENT
+ */
+#define EX_NSEL_COMMON	37
+typedef struct tpl_ext_37_s {
+	union {
+		uint64_t	event_time;
+		uint32_t	v[2];
+	};
+	uint32_t	conn_id;
+	union {
+		struct {
+#ifdef WORDS_BIGENDIAN
+			uint8_t		icmp_type;
+			uint8_t		icmp_code;
+#else
+			uint8_t		icmp_code;
+			uint8_t		icmp_type;
+#endif
+		};
+		uint16_t nsel_icmp;
+	};
+	uint8_t		fw_event;
+	uint8_t		fill;
+	uint16_t	fw_xevent;
+	uint16_t	fill2;
+	uint8_t		data[4];	// points to further data
+} tpl_ext_37_t;
 
-#define EX_NEL_0		42
-#define EX_NEL_1		43
-#define EX_NEL_2		44
+/*
+ * NSEL/NEL xlate ports
+ * +----+--------------+--------------+--------------+--------------+
+ * |  0 |  NF_F_XLATE_SRC_PORT(227)   |  NF_F_XLATE_DST_PORT(228)   |
+ * +----+--------------+--------------+--------------+--------------+
+ * ASA 8.4 compatibility mapping 40003 -> 227
+ * ASA 8.4 compatibility mapping 40004 -> 228
+ */
+#define EX_NSEL_XLATE_PORTS	38
+typedef struct tpl_ext_38_s {
+	uint16_t	xlate_src_port;
+	uint16_t	xlate_dst_port;
+	uint8_t		data[4];	// points to further data
+} tpl_ext_38_t;
+
+/*
+ * NSEL xlate v4 IP address
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  0 |                NF_F_XLATE_SRC_ADDR_IPV4(225)              |                NF_F_XLATE_DST_ADDR_IPV4(226)              |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * ASA 8.4 compatibility mapping 40001 -> 225
+ * ASA 8.4 compatibility mapping 40002 -> 226
+ */
+#define EX_NSEL_XLATE_IP_v4	39
+typedef struct tpl_ext_39_s {
+	uint32_t	xlate_src_ip;
+	uint32_t	xlate_dst_ip;
+	uint8_t		data[4];	// points to further data
+} tpl_ext_39_t;
+
+/*
+ * NSEL xlate v6 IP address - not yet implemented by CISCO
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  0 |                                                         xlate src ip (281)                                            |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  1 |                                                         xlate src ip (281)                                            |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  2 |                                                         xlate dst ip (282)                                            |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  3 |                                                         xlate dst ip (282)                                            |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ */
+#define EX_NSEL_XLATE_IP_v6	40
+typedef struct tpl_ext_40_s {
+	uint64_t	xlate_src_ip[2];
+	uint64_t	xlate_dst_ip[2];
+	uint8_t		data[4];	// points to further data
+} tpl_ext_40_t;
+
+
+/*
+ * NSEL ACL ingress/egress acl ID
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  0 |                                            NF_F_INGRESS_ACL_ID(33000)                                                 |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  1 |                 NF_F_INGRESS_ACL_ID(33000)                |               NF_F_EGRESS_ACL_ID(33001)                   |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  2 |                                            NF_F_EGRESS_ACL_ID(33001)                                                  |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ */
+#define EX_NSEL_ACL		41
+typedef struct tpl_ext_41_s {
+	uint32_t	ingress_acl_id[3];
+	uint32_t	egress_acl_id[3];
+	uint8_t		data[4];	// points to further data
+} tpl_ext_41_t;
+
+/*
+ * NSEL ACL username
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  0 |                                                  NF_F_USERNAME(40000)                                                 |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  1 |                                                                                                                       |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  2 |                                                                                                                       |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ */
+#define EX_NSEL_USER	42
+typedef struct tpl_ext_42_s {
+	char		username[24];
+	uint8_t		data[4];	// points to further data
+} tpl_ext_42_t;
+
+/*
+ * NSEL ACL username max
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  0 |                                                  NF_F_USERNAME(40000)                                                 |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * | .. |                                                                                                                       |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  8 |                                                                                                                       |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ */
+#define EX_NSEL_USER_MAX 43
+typedef struct tpl_ext_43_s {
+	char		username[72];
+	uint8_t		data[4];	// points to further data
+} tpl_ext_43_t;
+
+
+#define EX_NSEL_RESERVED 44
 
 /*
  * nprobe extensions
@@ -947,12 +1133,57 @@ typedef struct tpl_ext_latency_s {
 	uint8_t		data[4];	// points to further data
 } tpl_ext_latency_t;
 
+/*
+ * NEL xlate ports
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  0 |NAT_EVENT(230)|     flags    |            fill             |                  NF_N_EGRESS_VRFID(235)                   |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  1 |                 NF_N_INGRESS_VRFID(234)                   |
+ * +----+--------------+--------------+--------------+--------------+
+ */
+#define EX_NEL_COMMON 46
+typedef struct tpl_ext_46_s {
+	uint8_t		nat_event;
+	uint8_t		flags;
+	uint16_t	fill;
+	uint32_t	egress_vrfid;
+	uint32_t	ingress_vrfid;
+	uint8_t		data[4];	// points to further data
+} tpl_ext_46_t;
+
+#define EX_NEL_GLOBAL_IP_v4	47
+/* 
+ * no longer used. Mapped to NSEL extension EX_NSEL_XLATE_IP_v4
+ */
+typedef struct tpl_ext_47_s {
+	uint32_t    nat_inside;
+	uint32_t    nat_outside;
+	uint8_t     data[4];    // points to further data
+} tpl_ext_47_t;
+
+/*
+ * NEL Port Block Allocation
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  0 | NF_F_XLATE_PORT_BLOCK_START |  NF_F_XLATE_PORT_BLOCK_END  |  NF_F_XLATE_PORT_BLOCK_STEP |  NF_F_XLATE_PORT_BLOCK_SIZE |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ */
+#define EX_PORT_BLOCK_ALLOC	48
+typedef struct tpl_ext_48_s {
+	uint16_t	block_start;
+	uint16_t	block_end;
+	uint16_t	block_step;
+	uint16_t	block_size;
+	uint8_t     data[4];    // points to further data
+} tpl_ext_48_t;
+
+#define EX_NEL_RESERVED_1	49
+
 
 /* 
  * 
  * 
- * Extension map:
- * =============
+ * V1 Extension map:
+ * =================
  * The extension map replaces the individual flags in v1 layout. With many possible extensions and combination of extensions
  * an extension map is more efficient and flexible while reading and decoding the record.
  * In current version of nfdump, up to 65535 individual extension maps are supported, which is considered to be enough.
@@ -991,8 +1222,60 @@ typedef struct extension_map_s {
 } extension_map_t;
 
 
+/* 
+ * 
+ * 
+ * V2 Extension map:
+ * =================
+ * The V2 extension map replaces the V1 extension map. The basic extension architecture remains the same. V2 extensions
+ * adds extra information about the extension such as size and offset within the packed record. This allows more 
+ * flexibility by using flexible length extensions up to the extension defined maximum. With the introduction of V2
+ * extension maps, the old master_record will become obsolete in near future.
+ * Implementation wise: if extension size is set to 0 - the first 2 bytes of the extension data contains the size of
+ * the extension. Extensions with flexible size must only appended at the end of the record block.
+ * 
+ * For each available extension record, the ids are recorded in the extension map in the order they appear.
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  - |	     0     |      1       |      2       |      3       |      4       |      5       |      6       |      7       |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  0 |       record type == 10     |             size            |            map id           |      max extension size     |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  1 |       extension id 1        |   offset extension id 1     |        extension id 2       |    offset extension id 2    | 
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  1 |       extension id 3        |   offset extension id 3     |        extension id 4       |    offset extension id 4    | 
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * ...
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ * |  n |       extension id n        |      size extension id n    |              0              |              0              |
+ * +----+--------------+--------------+--------------+--------------+--------------+--------------+--------------+--------------+
+ */
+
+/* extension IDs above are 16 bit integers. So themax number of available extensions is */
+#define MAX_EXTENSIONS 65536
+/* XXX extension 1.7
+typedef struct extension_map17_s {
+ 	// record head
+ 	uint16_t	type;	// is ExtensionMapType2
+ 	uint16_t	size;	// size of full map incl. header
+
+	// map data
+	uint16_t	map_id;			// identifies this map
+ 	uint16_t	extension_size; // max size of all extensions
+	struct ex_def_s {
+		uint16_t	id;		// extension id
+		uint16_t	offset;	// max extension size
+	} ex[1];				// array of all extensions
+} extension_map17_t;
+*/
+
 // see nfx.c - extension_descriptor
+#ifdef NSEL
+// Defaults for NSEL
+#define DefaultExtensions  "1,8,26,27,28,29,30,31"
+#else
+// Collector netflow defaults
 #define DefaultExtensions  "1,2"
+#endif
 
 /*
  * nfcapd writes an info stat record for each new exporter
@@ -1119,8 +1402,6 @@ typedef struct sampler_record_s {
 } sampler_record_t;
 
 
-// typedef struct master_record_s master_record_t;
-
 /* the master record contains all possible records unpacked */
 typedef struct master_record_s {
 	// common information from all netflow versions
@@ -1129,21 +1410,15 @@ typedef struct master_record_s {
 
 	uint16_t	type;			// index 0  0xffff 0000 0000 0000
 	uint16_t	size;			// index 0	0x0000'ffff'0000 0000
-	uint8_t		flags;			// index 0	0x0000'0000'ff00'0000
-	uint8_t		exporter_sysid;	// index 0	0x0000'0000'00ff'0000
+	uint16_t	flags;			// index 0	0x0000'0000'ffff'0000
 	uint16_t	ext_map;		// index 0	0x0000'0000'0000'ffff
 #	define OffsetRecordFlags 	0
-#	define OffsetExporterSysID	0
 #ifdef WORDS_BIGENDIAN
 #	define MaskRecordFlags  	0x00000000ff000000LL
-#	define MaskExporterSysID  	0x0000000000ff0000LL
 #	define ShiftRecordFlags 	24
-#	define ShiftExporterSysID 	16
 #else
 #	define MaskRecordFlags  	0x000000ff00000000LL
-#	define MaskExporterSysID  	0x0000ff0000000000LL
 #	define ShiftRecordFlags 	32
-#	define ShiftExporterSysID 	40
 #endif
 
 	//
@@ -1196,73 +1471,58 @@ typedef struct master_record_s {
 #	define ShiftTos  			56
 #endif
 
-	// extension 8
 	uint16_t	srcport;		// index 3	0xffff'0000'0000'0000
 	uint16_t	dstport;		// index 3  0x0000'ffff'0000'0000
+	uint16_t	exporter_sysid; // index 3	0x0000'0000'ffff'0000
+
 	union {
 		struct {
-			uint8_t	dst_tos;	// index 3  0x0000'0000'ff00'0000
-			uint8_t	dir;		// index 3  0x0000'0000'00ff'0000
-			uint8_t	src_mask;	// index 3  0x0000'0000'0000'ff00
-			uint8_t	dst_mask;	// index 3  0x0000'0000'0000'00ff
+#ifdef WORDS_BIGENDIAN
+			uint8_t		icmp_type;	// index 3  0x0000'0000'0000'ff00
+			uint8_t		icmp_code;	// index 3  0x0000'0000'0000'00ff
+#else
+			// little endian confusion ...
+			uint8_t		icmp_code;	
+			uint8_t		icmp_type;
+#endif
 		};
-		uint32_t	any;
+		uint16_t icmp;
 	};
+
+
 #ifdef WORDS_BIGENDIAN
 #	define OffsetPort 			3
+#	define OffsetExporterSysID	3
 #	define MaskSrcPort			0xffff000000000000LL
 #	define ShiftSrcPort			48
 
 #	define MaskDstPort			0x0000ffff00000000LL
 #	define ShiftDstPort 		32	
 
-#	define MaskICMPtype			0x0000ff0000000000LL
-#	define ShiftICMPtype 		40
-#	define MaskICMPcode			0x000000ff00000000LL
-#	define ShiftICMPcode 		32
+#	define MaskExporterSysID  	0x00000000ffff0000LL
+#	define ShiftExporterSysID 	16
 
-#	define OffsetDstTos			3
-#	define MaskDstTos			0x00000000ff000000LL
-#	define ShiftDstTos  		24
-
-#	define OffsetDir			3
-#	define MaskDir				0x0000000000ff0000LL
-#	define ShiftDir  			16
-
-#	define OffsetMask			3
-#	define MaskSrcMask			0x000000000000ff00LL
-#	define ShiftSrcMask  		8
-
-#	define MaskDstMask			0x00000000000000ffLL
-#	define ShiftDstMask 		0
+#	define MaskICMPtype			0x000000000000ff00LL
+#	define ShiftICMPtype 		8
+#	define MaskICMPcode			0x00000000000000ffLL
+#	define ShiftICMPcode 		0
 
 #else
 #	define OffsetPort 			3
+#	define OffsetExporterSysID	3
 #	define MaskSrcPort			0x000000000000ffffLL
 #	define ShiftSrcPort			0
 
 #	define MaskDstPort			0x00000000ffff0000LL
 #	define ShiftDstPort 		16
 
-#	define MaskICMPtype			0x00000000ff000000LL
-#	define ShiftICMPtype 		24
-#	define MaskICMPcode			0x0000000000ff0000LL
-#	define ShiftICMPcode 		16
+#	define MaskExporterSysID  	0x0000ffff00000000LL
+#	define ShiftExporterSysID 	32
 
-#	define OffsetDstTos			3
-#	define MaskDstTos			0x000000ff00000000LL
-#	define ShiftDstTos  		32
-
-#	define OffsetDir			3
-#	define MaskDir				0x0000ff0000000000LL
-#	define ShiftDir  			40
-
-#	define OffsetMask			3
-#	define MaskSrcMask			0x00ff000000000000LL
-#	define ShiftSrcMask 		48
-
-#	define MaskDstMask			0xff00000000000000LL
-#	define ShiftDstMask  		56
+#	define MaskICMPtype			0xff00000000000000LL
+#	define ShiftICMPtype 		56
+#	define MaskICMPcode			0x00ff000000000000LL
+#	define ShiftICMPcode 		48
 #endif
 
 	// extension 4 / 5
@@ -1327,6 +1587,9 @@ typedef struct master_record_s {
 			uint64_t	dstaddr[2];	// dstaddr[0-1] index 8 0xffff'ffff'ffff'ffff
 									// dstaddr[2-3] index 9 0xffff'ffff'ffff'ffff
 		} _v6;
+		struct _ip64_s {
+			uint64_t	addr[4];
+		} _ip_64;
 	} ip_union;
 
 #ifdef WORDS_BIGENDIAN
@@ -1420,26 +1683,67 @@ typedef struct master_record_s {
 #	define OffsetBGPNexthopv6b	15
 #endif
 
+	// extension 8
+	union {
+		struct {
+			uint8_t	dst_tos;	// index 16 0xff00'0000'0000'0000
+			uint8_t	dir;		// index 16 0x00ff'0000'0000'0000
+			uint8_t	src_mask;	// index 16 0x0000'ff00'0000'0000
+			uint8_t	dst_mask;	// index 16 0x0000'00ff'0000'0000
+		};
+		uint32_t	any;
+	};
+
 	// extension 13
-	uint16_t	src_vlan;		// index 16 0xffff'0000'0000'0000
-	uint16_t	dst_vlan;		// index 16 0x0000'ffff'0000'0000
-	uint32_t	fill1;			// align 64bit word
+	uint16_t	src_vlan;		// index 16 0x0000'0000'ffff'0000
+	uint16_t	dst_vlan;		// index 16 0x0000'0000'0000'ffff
 
 #ifdef WORDS_BIGENDIAN
-#	define OffsetVlan 			16	
-#	define MaskSrcVlan  		0xffff000000000000LL
-#	define ShiftSrcVlan 		48
+#	define OffsetDstTos			16
+#	define MaskDstTos			0xff00000000000000LL
+#	define ShiftDstTos  		56
 
-#	define MaskDstVlan  		0x0000ffff00000000LL
-#	define ShiftDstVlan 		32
+#	define OffsetDir			16
+#	define MaskDir				0x00ff000000000000LL
+#	define ShiftDir  			48
+
+#	define OffsetMask			16
+#	define MaskSrcMask			0x0000ff0000000000LL
+#	define ShiftSrcMask  		40
+
+#	define MaskDstMask			0x000000ff00000000LL
+#	define ShiftDstMask 		32
+
+#	define OffsetVlan 			16	
+#	define MaskSrcVlan  		0x00000000ffff0000LL
+#	define ShiftSrcVlan 		16
+
+#	define MaskDstVlan  		0x000000000000ffffLL
+#	define ShiftDstVlan 		0
 
 #else
+#	define OffsetDstTos			16
+#	define MaskDstTos			0x00000000000000ffLL
+#	define ShiftDstTos  		0
+
+#	define OffsetDir			16
+#	define MaskDir				0x000000000000ff00LL
+#	define ShiftDir  			8
+
+#	define OffsetMask			16
+#	define MaskSrcMask			0x0000000000ff0000LL
+#	define ShiftSrcMask  		16
+
+#	define MaskDstMask			0x00000000ff000000LL
+#	define ShiftDstMask 		24
+
 #	define OffsetVlan 			16	
-#	define MaskSrcVlan  		0x000000000000ffffLL
-#	define ShiftSrcVlan 		0
+#	define MaskSrcVlan  		0x0000ffff00000000LL
+#	define ShiftSrcVlan 		32
+
+#	define MaskDstVlan  		0xffff000000000000LL
+#	define ShiftDstVlan 		48
 
-#	define MaskDstVlan  		0x00000000ffff0000LL
-#	define ShiftDstVlan 		16
 #endif
 
 	// extension 14 / 15
@@ -1565,13 +1869,188 @@ typedef struct master_record_s {
 #	define ShiftBGPadjNext		0
 #	define MaskBGPadjPrev		0xFFFFFFFF00000000LL
 #	define ShiftBGPadjPrev		32
+#endif
+
+	// NSEL extensions
+#ifdef NSEL 
+#define NSEL_BASE_OFFSET     (offsetof(master_record_t, conn_id) >> 3)
+
+	// common block
+#   define OffsetConnID  NSEL_BASE_OFFSET
+#   define OffsetNATevent  NSEL_BASE_OFFSET
+	uint32_t	conn_id;			// index OffsetConnID    0xffff'ffff'0000'0000
+	uint8_t		event;				// index OffsetConnID    0x0000'0000'ff00'0000
+#define FW_EVENT 1
+#define NAT_EVENT 2
+	uint8_t		event_flag;			// index OffsetConnID    0x0000'0000'00ff'0000
+	uint16_t	fw_xevent;			// index OffsetConnID    0x0000'0000'0000'ffff
+	uint64_t	event_time;			// index OffsetConnID +1 0x1111'1111'1111'1111
+#ifdef WORDS_BIGENDIAN
+#	define MaskConnID		0xFFFFFFFF00000000LL
+#	define ShiftConnID		32
+#	define MaskFWevent		0x00000000FF000000LL
+#	define ShiftFWevent		24
+#	define MasNATevent		0x00000000FF000000LL
+#	define ShiftNATevent	24
+#	define MaskFWXevent		0x000000000000FFFFLL
+#	define ShiftFWXevent	0
+#else
+#	define MaskConnID		0x00000000FFFFFFFFLL
+#	define ShiftConnID		0
+#	define MaskFWevent		0x000000FF00000000LL
+#	define ShiftFWevent		32
+#	define MasNATevent		0x000000FF00000000LL
+#	define ShiftNATevent	32
+#	define MaskFWXevent		0xFFFF000000000000LL
+#	define ShiftFWXevent	48
+
+#endif
+
+	// xlate ip/port
+#   define OffsetXLATEPort NSEL_BASE_OFFSET+2
+	uint16_t	xlate_src_port;		// index OffsetXLATEPort 0xffff'0000'0000'0000
+	uint16_t	xlate_dst_port;		// index OffsetXLATEPort 0x0000'ffff'0000'0000
+	uint32_t	xlate_flags;
+#   define OffsetXLATESRCIP NSEL_BASE_OFFSET+3
+	ip_addr_t	xlate_src_ip;		// ipv4  OffsetXLATESRCIP +1 0x0000'0000'ffff'ffff
+									// ipv6	 OffsetXLATESRCIP 	 0xffff'ffff'ffff'ffff
+									// ipv6	 OffsetXLATESRCIP	 0xffff'ffff'ffff'ffff
+
+	ip_addr_t	xlate_dst_ip;		// ipv4  OffsetXLATEDSTIP +1 0x0000'0000'ffff'ffff
+									// ipv6	 OffsetXLATEDSTIP 	 0xffff'ffff'ffff'ffff
+									// ipv6	 OffsetXLATEDSTIP 	 0xffff'ffff'ffff'ffff
+#ifdef WORDS_BIGENDIAN
+#	define MaskXLATESRCPORT	 0xFFFF000000000000LL
+#	define ShiftXLATESRCPORT 48
+#	define MaskXLATEDSTPORT	 0x0000FFFF00000000LL
+#	define ShiftXLATEDSTPORT 32
+
+#	define OffsetXLATESRCv4	 OffsetXLATESRCIP+1
+#	define MaskXLATEIPv4  	 0x00000000fFFFFFFFLL
+#	define ShiftXLATEIPv4 	 0
+
+#	define OffsetXLATESRCv6a OffsetXLATESRCIP
+#	define OffsetXLATESRCv6b OffsetXLATESRCIP+1
+
+#	define OffsetXLATEDSTv6a OffsetXLATESRCIP+2
+#	define OffsetXLATEDSTv6b OffsetXLATESRCIP+3
+
+#else
+#	define MaskXLATESRCPORT	 0x000000000000FFFFLL
+#	define ShiftXLATESRCPORT 0
+#	define MaskXLATEDSTPORT	 0x00000000FFFF0000LL
+#	define ShiftXLATEDSTPORT 16
+
+#	define OffsetXLATESRCv4	 OffsetXLATESRCIP+1
+#	define MaskXLATEIPv4  	 0xFFFFFFFF00000000LL
+#	define ShiftXLATEIPv4 	 32
+
+#	define OffsetXLATESRCv6a OffsetXLATESRCIP
+#	define OffsetXLATESRCv6b OffsetXLATESRCIP+1
+
+#	define OffsetXLATEDSTv6a OffsetXLATESRCIP+2
+#	define OffsetXLATEDSTv6b OffsetXLATESRCIP+3
+
+#endif
+
+
+	// ingress/egress ACL id
+#   define OffsetIngressAclId NSEL_BASE_OFFSET+7
+#	define OffsetIngressAceId NSEL_BASE_OFFSET+7
+#	define OffsetIngressGrpId NSEL_BASE_OFFSET+8
+#	define OffsetEgressAclId  NSEL_BASE_OFFSET+8
+#	define OffsetEgressAceId  NSEL_BASE_OFFSET+9
+#	define OffsetEgressGrpId  NSEL_BASE_OFFSET+9
+	uint32_t ingress_acl_id[3];	// index OffsetIngressAclId   0xffff'ffff'0000'0000
+								// index OffsetIngressAceId   0x0000'0000'ffff'ffff
+								// index OffsetIngressGrpId   0xffff'ffff'0000'0000
+	uint32_t egress_acl_id[3];	// index OffsetEgressAclId	  0x0000'0000'ffff'ffff
+								// index OffsetEgressAceId	  0xffff'ffff'0000'0000
+								// index OffsetEgressGrpId	  0x0000'0000'ffff'ffff
+#ifdef WORDS_BIGENDIAN
+#define MaskIngressAclId	0xffffffff00000000LL
+#define ShiftIngressAclId	32
+#define MaskIngressAceId	0x00000000ffffffffLL
+#define ShiftIngressAceId	0
+#define MaskIngressGrpId	0xffffffff00000000LL
+#define ShiftIngressGrpId	32
+#define MaskEgressAclId		0x00000000ffffffffLL
+#define ShiftEgressAclId	0
+#define MaskEgressAceId		0xffffffff00000000LL
+#define ShiftEgressAceId	32
+#define MaskEgressGrpId		0x00000000ffffffffLL
+#define ShiftEgressGrpId	0
+#else
+#define MaskIngressAclId	0x00000000ffffffffLL
+#define ShiftIngressAclId	0
+#define MaskIngressAceId	0xffffffff00000000LL
+#define ShiftIngressAceId	32
+#define MaskIngressGrpId	0x00000000ffffffffLL
+#define ShiftIngressGrpId	0
+#define MaskEgressAclId		0xffffffff00000000LL
+#define ShiftEgressAclId	32
+#define MaskEgressAceId		0x00000000ffffffffLL
+#define ShiftEgressAceId	0
+#define MaskEgressGrpId		0xffffffff00000000LL
+#define ShiftEgressGrpId	32
+#endif
+
+	// username
+#	define OffsetUsername  NSEL_BASE_OFFSET+10
+	char username[72];
+
+	// NAT extensions
+	// NAT event is mapped into ASA event
+#define NAT_BASE_OFFSET     (offsetof(master_record_t, ingress_vrfid) >> 3)
+	// common block
+#   define OffsetNELcommon  NEL_BASE_OFFSET
+#   define OffsetIVRFID  	NAT_BASE_OFFSET
+#   define OffsetEVRFID  	NAT_BASE_OFFSET
+#   define OffsetPortBlock	NAT_BASE_OFFSET+1
+	uint32_t	ingress_vrfid;	// OffsetIVRFID	   0xffff'ffff'0000'0000
+	uint32_t	egress_vrfid;	// OffsetEVRFID	   0x0000'0000'ffff'ffff
+
+	// Port block allocation
+	uint16_t	block_start;	// OffsetPortBlock 0xffff'0000'0000'0000
+	uint16_t	block_end;		// OffsetPortBlock 0x0000'ffff'0000'0000
+	uint16_t	block_step;		// OffsetPortBlock 0x0000'0000'ffff'0000
+	uint16_t	block_size;		// OffsetPortBlock 0x0000'0000'0000'ffff
+
+#ifdef WORDS_BIGENDIAN
+#	define MaskIVRFID			0xFFFFFFFF00000000LL
+#	define ShiftIVRFID			32
+#	define MaskEVRFID			0x00000000FFFFFFFFLL
+#	define ShiftEVRFID			0
+#	define MaskPortBlockStart	0xFFFF000000000000LL
+#	define ShiftPortBlockStart	48
+#	define MaskPortBlockEnd		0x0000FFFF00000000LL
+#	define ShiftPortBlockEnd	32
+#	define MaskPortBlockStep	0x00000000FFFF0000LL
+#	define ShiftPortBlockStep	16
+#	define MaskPortBlockSize	0x000000000000FFFFLL
+#	define ShiftPortBlockSize	0
+#else
+#	define MaskIVRFID			0x00000000FFFFFFFFLL
+#	define ShiftIVRFID			0
+#	define MaskEVRFID			0xFFFFFFFF00000000LL
+#	define ShiftEVRFID			32
+#	define MaskPortBlockStart	0x000000000000FFFFLL
+#	define ShiftPortBlockStart	0
+#	define MaskPortBlockEnd		0x00000000FFFF0000LL
+#	define ShiftPortBlockEnd	16
+#	define MaskPortBlockStep	0x0000FFFF00000000LL
+#	define ShiftPortBlockStep	32
+#	define MaskPortBlockSize	0xFFFF000000000000LL
+#	define ShiftPortBlockSize	48
+#endif
+
 #endif
 
 	// nprobe extensions
 	// latency extension
-	uint64_t	client_nw_delay_usec;	// index 32 0xffff'ffff'ffff'ffff
-	uint64_t	server_nw_delay_usec;	// index 33 0xffff'ffff'ffff'ffff
-	uint64_t	appl_latency_usec;		// index 34 0xffff'ffff'ffff'ffff
+	uint64_t	client_nw_delay_usec;	// index LATENCY_BASE_OFFSET 0xffff'ffff'ffff'ffff
+	uint64_t	server_nw_delay_usec;	// index LATENCY_BASE_OFFSET + 1 0xffff'ffff'ffff'ffff
+	uint64_t	appl_latency_usec;		// index LATENCY_BASE_OFFSET + 2 0xffff'ffff'ffff'ffff
 
 #define LATENCY_BASE_OFFSET     (offsetof(master_record_t, client_nw_delay_usec) >> 3)
 #   define OffsetClientLatency  LATENCY_BASE_OFFSET
@@ -1681,7 +2160,11 @@ typedef struct common_record_v1_s {
 
 
 // a few handy shortcuts
-#define FILE_IS_COMPRESSED(n) ((n)->file_header->flags & FLAG_COMPRESSED)
+#define FILE_IS_LZO_COMPRESSED(n) ((n)->file_header->flags & FLAG_LZO_COMPRESSED)
+#define FILE_IS_BZ2_COMPRESSED(n) ((n)->file_header->flags & FLAG_BZ2_COMPRESSED)
+#define FILE_COMPRESSION(n) ( FILE_IS_LZO_COMPRESSED(n) ? LZO_COMPRESSED : FILE_IS_BZ2_COMPRESSED(n) ? BZ2_COMPRESSED : NOT_COMPRESSED )
+#define FILE_IS_NOT_COMPRESSED(n) ( (FILE_IS_LZO_COMPRESSED(n) + FILE_IS_BZ2_COMPRESSED(n)) == 0 )
+
 #define BLOCK_IS_COMPRESSED(n) ((n)->flags == 2 )
 #define HAS_CATALOG(n) ((n)->file_header->flags & FLAG_CATALOG)
 #define IP_ANONYMIZED(n) ((n)->file_header->flags & FLAG_ANONYMIZED)
@@ -1690,7 +2173,7 @@ void SumStatRecords(stat_record_t *s1, stat_record_t *s2);
 
 nffile_t *OpenFile(char *filename, nffile_t *nffile);
 
-nffile_t *OpenNewFile(char *filename, nffile_t *nffile, int compressed, int anonymized, char *ident);
+nffile_t *OpenNewFile(char *filename, nffile_t *nffile, int compress, int anonymized, char *ident);
 
 nffile_t *AppendFile(char *filename);
 
@@ -1714,7 +2197,9 @@ int WriteBlock(nffile_t *nffile);
 
 int WriteExtraBlock(nffile_t *nffile, data_block_header_t *block_header);
 
-void UnCompressFile(char * filename);
+int RenameAppend(char *from, char *to);
+
+void ModifyCompressFile(char * rfile, char *Rfile, int compress);
 
 void ExpandRecord_v1(common_record_t *input_record,master_record_t *output_record );
 
